# -*- coding: utf-8 -*-
import os
import time
import uuid
import copy
from common import Constant
from multiprocessing import Process, Queue
from loguru import logger
from enums.AnalysisStatusEnum import AnalysisStatus
from enums.AnalysisTypeEnum import AnalysisType
from enums.ExceptionEnum import ExceptionType
from enums.ModelTypeEnum import ModelType
from util import LogUtils, TimeUtils
from util.Cv2Utils import Cv2Util
from util import ModelUtils
from entity.FeedBack import message_feedback
from util import AliyunSdk
from concurrency.CommonThread import Common
from concurrency.CommonProcess import CommonProcess
from exception.CustomerException import ServiceException


class IntelligentRecognitionProcess(Process):
    def __init__(self, fbQueue, content, msg, imageQueue, gpu_ids):
        super().__init__()
        self.fbQueue = fbQueue
        self.eventQueue = Queue()
        self.content = content
        self.msg = msg
        self.imageQueue = imageQueue
        self.gpu_ids = gpu_ids

    # 给本进程发送事件
    def sendEvent(self, eBody):
        self.eventQueue.put(eBody)

    # 获取下一个事件
    def getEvent(self):
        try:
            eBody = self.eventQueue.get(block=False)
            return eBody
        except Exception as e:
            pass

    # 推送执行结果
    def sendResult(self, result):
        self.fbQueue.put(result)


    '''
    "models": [{
        "code": "模型编号",
        "categories":[{
            "id": "模型id",
            "config": {
                "k1": "v1",
                "k2": "v2"
            }
        }]
    }]
    '''

    def get_model(self, gpuId, models):
        for model in models:
            try:
                code = model.get("code")
                needed_objectsIndex = [int(category.get("id")) for category in model.get("categories")]
                if code == ModelType.WATER_SURFACE_MODEL.value[1]:
                    return ModelUtils.SZModel(gpuId, needed_objectsIndex), code
                elif code == ModelType.FOREST_FARM_MODEL.value[1]:
                    return ModelUtils.LCModel(gpuId, needed_objectsIndex), code
                else:
                    logger.error("未匹配到对应的模型")
                    raise ServiceException(ExceptionType.AI_MODEL_MATCH_EXCEPTION.value[0],
                                           ExceptionType.AI_MODEL_MATCH_EXCEPTION.value[1])
            except Exception as e:
                logger.error("获取模型配置异常：")
                logger.exception(e)
                raise ServiceException(ExceptionType.AI_MODEL_CONFIG_EXCEPTION.value[0],
                                       ExceptionType.AI_MODEL_CONFIG_EXCEPTION.value[1])


'''
    实时任务进程
'''


class OnlineIntelligentRecognitionProcess(IntelligentRecognitionProcess):

    # 停止任务方法
    def stop_task(self, cv2tool, orFilePath, aiFilePath, commonProcess, snalysisStatus):
        # 停止cv2相关配置
        cv2tool.stop_cv2()
        if not os.path.exists(orFilePath) or not os.path.exists(aiFilePath):
            logger.error("原视频或AI视频不存在！requestId:{}", self.msg.get("request_id"))
            raise ServiceException(ExceptionType.VIDEO_ADDRESS_EXCEPTION.value[0],
                                   ExceptionType.VIDEO_ADDRESS_EXCEPTION.value[1])
        params1 = (orFilePath, "orOnLineVideo", self.content)
        or_update_thread = Common(content=self.content, func=AliyunSdk.get_play_url, args=params1)
        params2 = (aiFilePath, "aiOnLineVideo", self.content)
        ai_update_thread = Common(content=self.content, func=AliyunSdk.get_play_url, args=params2)
        or_update_thread.start()
        ai_update_thread.start()
        or_play_url = or_update_thread.get_result()
        ai_play_url = ai_update_thread.get_result()
        if or_play_url is None or ai_play_url is None:
            logger.error("原视频或AI视频播放上传VOD失败！原视频播放地址：{}, AI播放地址: {}, requestId: {}", or_play_url,
                         ai_play_url, self.msg.get("request_id"))
            raise ServiceException(ExceptionType.GET_VIDEO_URL_EXCEPTION.value[0],
                                   ExceptionType.GET_VIDEO_URL_EXCEPTION.value[1])
        commonProcess.sendEvent({"command": "stop_heartbeat_imageFileUpdate"})
        self.sendResult(message_feedback(self.msg.get("request_id"), snalysisStatus,
                                         AnalysisType.ONLINE.value,
                                         progress=Constant.success_progess,
                                         original_url=or_play_url,
                                         sign_url=ai_play_url,
                                         analyse_time=TimeUtils.now_date_to_str()))

    def run(self):
        LogUtils.init_log(self.content)
        # 程序开始时间
        start_time = time.time()
        # 启动公共进程包含（图片上传线程，心跳线程，问题反馈线程）
        commonProcess = CommonProcess(self.fbQueue, None, self.content, self.msg, self.imageQueue,
                                      AnalysisType.ONLINE.value)
        commonProcess.start()
        orFilePath = None
        aiFilePath = None
        cv2tool = None
        try:
            # 加载模型
            logger.info("开始加载算法模型, requestId: {}", self.msg.get("request_id"))
            mod, model_type_code = self.get_model(str(self.gpu_ids[0]), self.msg["models"])
            logger.info("加载算法模型完成, requestId: {}", self.msg.get("request_id"))
            # 定义原视频、AI视频保存名称
            randomStr = str(uuid.uuid1().hex)
            orFilePath = os.path.join(self.content["video"]["file_path"],
                                      randomStr + "_or_" + self.msg.get("request_id") + ".mp4")
            aiFilePath = os.path.join(self.content["video"]["file_path"],
                                      randomStr + "_ai_" + self.msg.get("request_id") + ".mp4")
            logger.info("开始初始化cv2, requestId: {}", self.msg.get("request_id"))
            cv2tool = Cv2Util(self.msg.get('pull_url'), self.msg.get('push_url'), orFilePath, aiFilePath)
            cv2tool.build_cv2()
            # cv2重试初始化次数
            cv2_init_num = 1
            # 解决开始拉流失败问题，解决初始化fps,height,weight获取不到问题
            while True:
                if cv2tool.cap is None or not cv2tool.cap.isOpened():
                    pull_stream_init_timeout = time.time()
                    if pull_stream_init_timeout - start_time > int(self.content["service"]["cv2_pull_stream_timeout"]):
                        logger.info("{}: 开始拉流超时, 超时时间：{}", self.msg.get("request_id"),
                                    int(pull_stream_init_timeout - start_time))
                        raise ServiceException(ExceptionType.PULLSTREAM_TIMEOUT_EXCEPTION.value[0],
                                               ExceptionType.PULLSTREAM_TIMEOUT_EXCEPTION.value[1])
                    logger.info("cv2初始化重试：{}次, requestId: {}", cv2_init_num, self.msg.get("request_id"))
                    cv2_init_num += 1
                    time.sleep(2)
                    cv2tool.build_cv2()
                    continue
                else:
                    break
            cv2tool.build_write()
            logger.info("初始化cv2完成, requestId: {}", self.msg.get("request_id"))
            # 当时视频帧数
            current_frame = 1
            high_score_image = {}
            step = int(self.content["service"]["frame_step"])
            pull_start_time = None
            read_start_time = None
            while True:
                end_time = time.time()
                create_task_time = int(end_time - start_time)
                if create_task_time > int(self.content["service"]["timeout"]):
                    logger.error("分析超时, 超时时间：{}, requestId: {}", create_task_time, self.msg.get("request_id"))
                    raise ServiceException(ExceptionType.ANALYSE_TIMEOUT_EXCEPTION.value[0],
                                           ExceptionType.ANALYSE_TIMEOUT_EXCEPTION.value[1])
                if not commonProcess.is_alive():
                    logger.info("图片上传、心跳、问题反馈进程停止异常, requestId: {}", self.msg.get("request_id"))
                    raise Exception("图片上传、心跳、问题反馈进程异常停止")
                eBody = self.getEvent()
                if eBody is not None and len(eBody) > 0:
                    logger.info("eBody:{} 任务停止分析, requestId: {}", eBody, self.msg.get("request_id"))
                    cmdStr = eBody.get("command", None)
                    logger.info("cmdStr:{} 任务停止分析, requestId: {}", cmdStr, self.msg.get("request_id"))
                    # 接收到停止指令
                    if 'stop' == cmdStr:
                        logger.info("{} 任务停止分析", self.msg.get("request_id"))
                        self.stop_task(cv2tool, orFilePath, aiFilePath, commonProcess, AnalysisStatus.SUCCESS.value)
                        break
                # 检测是否断流
                if cv2tool.cap is None or not cv2tool.cap.isOpened():
                    if pull_start_time is None:
                        pull_start_time = time.time()
                    # 默认1个小时
                    pull_stream_timeout = time.time()
                    if pull_stream_timeout - pull_start_time > int(self.content["service"]["cv2_pull_stream_timeout"]):
                        logger.info("{}: 开始拉流超时, 超时时间：{}", self.msg.get("request_id"),
                                    int(pull_stream_timeout - pull_start_time))
                        raise ServiceException(ExceptionType.PULLSTREAM_TIMEOUT_EXCEPTION.value[0],
                                               ExceptionType.PULLSTREAM_TIMEOUT_EXCEPTION.value[1])
                    logger.info("cv2初始化重试：{}次, requestId: {}", cv2_init_num, self.msg.get("request_id"))
                    cv2_init_num += 1
                    time.sleep(2)
                    cv2tool.build_cv2()
                    continue
                pull_start_time = None
                is_opened, frame = cv2tool.getCap().read()
                if is_opened is None or not is_opened:
                    if read_start_time is None:
                        read_start_time = time.time()
                    read_stream_timeout = time.time()
                    if read_stream_timeout - read_start_time > int(self.content["service"]["cv2_read_stream_timeout"]):
                        logger.info("{}: 运行中拉流超时, 超时时间：{}", self.msg.get("request_id"),
                                    int(read_stream_timeout - read_start_time))
                        self.stop_task(cv2tool, orFilePath, aiFilePath, commonProcess, AnalysisStatus.TIMEOUT.value)
                        break
                    time.sleep(2)
                    logger.info("cv2初始化重试：{}次, requestId: {}", cv2_init_num, self.msg.get("request_id"))
                    cv2_init_num += 1
                    cv2tool.build_cv2()
                    continue
                read_start_time = None
                # time00 = time.time()
                # 调用AI模型
                p_result, timeOut = mod.process(copy.deepcopy(frame))
                # time11 = time.time()
                # logger.info("算法模型调度时间：{}s", int(time11-time00))
                # AI推流
                try:
                    cv2tool.getP().stdin.write(p_result[1].tostring())
                    cv2tool.getOrVideoFile().write(frame)
                    frame_merge = cv2tool.video_merge(copy.deepcopy(frame), copy.deepcopy(p_result[1]))
                    cv2tool.getAiVideoFile().write(frame_merge)
                except Exception as e:
                    logger.error("requestId:{}, 写流异常：", self.msg.get("request_id"))
                    logger.exception(e)
                # # 问题图片加入队列, 暂时写死，后期修改为真实问题
                if p_result[2] is not None and len(p_result[2]) > 0:
                    ai_analyse_results = p_result[2]
                    for ai_analyse_result in ai_analyse_results:
                        order = str(int(ai_analyse_result[0]))
                        high_result = high_score_image.get(order)
                        conf_c = ai_analyse_result[5]
                        if high_result is None:
                            high_score_image[order] = {
                                "or_frame": frame,
                                "ai_frame": p_result[1],
                                "current_frame": current_frame,
                                "last_frame": current_frame + step,
                                "progress": "",
                                "mode_service": "online",
                                "model_type_code": model_type_code,
                                "model_detection_code": order,
                                "socre": conf_c
                            }
                        else:
                            if conf_c > high_result.get("socre"):
                                high_score_image[order] = {
                                    "or_frame": frame,
                                    "ai_frame": p_result[1],
                                    "current_frame": current_frame,
                                    "last_frame": current_frame + step,
                                    "progress": "",
                                    "mode_service": "online",
                                    "model_type_code": model_type_code,
                                    "model_detection_code": order,
                                    "socre": conf_c
                                }
                if current_frame % int(self.content["service"]["frame_step"]) == 0 and len(high_score_image) > 0:
                    # 2 [[2.0, 0.3059895932674408, 0.8865740895271301, 0.0338541679084301, 0.21759259700775146, 0.5478515625]]
                    #       其中每一个元素表示一个目标构成如：[float(cls_c), xc,yc,w,h, float(conf_c)]
                    #       #cls_c--类别，如0,1,2,3;  xc,yc,w,h--中心点坐标及宽；conf_c--得分, 取值范围在0-1之间
                    for value in high_score_image.values():
                        self.imageQueue.put(value)
                    high_score_image.clear()
                current_frame += 1
            logger.info("实时进程任务完成，requestId:{}", self.msg.get("request_id"))
        except ServiceException as s:
            commonProcess.sendEvent({"command": "stop_heartbeat_imageFileUpdate"})
            logger.error("服务异常，异常编号:{}, 异常描述：{}, requestId: {}", s.code, s.msg, self.msg.get("request_id"))
            self.sendResult(message_feedback(self.msg.get("request_id"), AnalysisStatus.FAILED.value,
                                             AnalysisType.ONLINE.value,
                                             s.code,
                                             s.msg,
                                             analyse_time=TimeUtils.now_date_to_str()))
        except Exception as e:
            commonProcess.sendEvent({"command": "stop_heartbeat_imageFileUpdate"})
            logger.error("requestId: {}, 未知异常：", self.msg.get("request_id"))
            logger.exception(e)
            self.sendResult(message_feedback(self.msg.get("request_id"), AnalysisStatus.FAILED.value,
                                             AnalysisType.ONLINE.value,
                                             ExceptionType.SERVICE_INNER_EXCEPTION.value[0],
                                             ExceptionType.SERVICE_INNER_EXCEPTION.value[1],
                                             analyse_time=TimeUtils.now_date_to_str()))
        finally:
            if cv2tool is not None:
                cv2tool.stop_cv2()
            time.sleep(3)
            commonProcess.sendEvent({"command": "stop"})
            commonProcess.join()
            # 删除本地视频文件
            if orFilePath is not None and os.path.exists(orFilePath):
                logger.info("开始删除原视频, orFilePath: {}, requestId: {}", orFilePath, self.msg.get("request_id"))
                os.remove(orFilePath)
                logger.info("删除原视频成功, orFilePath: {}, requestId: {}", orFilePath, self.msg.get("request_id"))
            if aiFilePath is not None and os.path.exists(aiFilePath):
                logger.info("开始删除AI视频, aiFilePath: {}, requestId: {}", aiFilePath, self.msg.get("request_id"))
                os.remove(aiFilePath)
                logger.info("删除AI视频成功, aiFilePath: {}, requestId: {}", aiFilePath, self.msg.get("request_id"))


class OfflineIntelligentRecognitionProcess(IntelligentRecognitionProcess):

    def stop_task(self, cv2tool, aiFilePath, commonProcess, analysisStatus):
        cv2tool.stop_cv2()
        if not os.path.exists(aiFilePath):
            logger.error("AI视频不存在！requestId:{}", self.msg.get("request_id"))
            raise ServiceException(ExceptionType.VIDEO_ADDRESS_EXCEPTION.value[0],
                                   ExceptionType.VIDEO_ADDRESS_EXCEPTION.value[1])
        params2 = (aiFilePath, "aiOffLineVideo", self.content)
        ai_update_thread = Common(content=self.content, func=AliyunSdk.get_play_url, args=params2)
        ai_update_thread.start()
        ai_play_url = ai_update_thread.get_result()
        if ai_play_url is None:
            logger.error("原视频或AI视频播放上传VOD失败！requestId: {}, AI播放地址: {}",
                         self.msg.get("request_id"), ai_play_url)
            raise ServiceException(ExceptionType.GET_VIDEO_URL_EXCEPTION.value[0],
                                   ExceptionType.GET_VIDEO_URL_EXCEPTION.value[1])
        # (requestId, status, type, error_code="", error_msg="",  progress="", original_url="", sign_url="",
        # category_id="", description="", analyse_time="")
        commonProcess.sendEvent({"command": "stop_heartbeat_imageFileUpdate"})
        self.sendResult(message_feedback(self.msg.get("request_id"), analysisStatus,
                                         AnalysisType.OFFLINE.value,
                                         progress=Constant.success_progess,
                                         sign_url=ai_play_url,
                                         analyse_time=TimeUtils.now_date_to_str()))

    def run(self):
        LogUtils.init_log(self.content)
        # 程序开始时间
        start_time = time.time()
        # 创建心跳队列
        hbQueue = Queue()
        # 结果反馈进程启动
        commonProcess = CommonProcess(self.fbQueue, hbQueue, self.content, self.msg, self.imageQueue,
                                      AnalysisType.OFFLINE.value)
        commonProcess.start()
        aiFilePath = None
        cv2tool = None
        try:
            # 加载模型
            logger.info("开始加载算法模型, requestId:{}", self.msg.get("request_id"))
            mod, model_type_code = self.get_model(str(self.gpu_ids[0]), self.msg["models"])
            # mod = ModelUtils.SZModel([0,1,2,3])
            logger.info("加载算法模型完成, requestId:{}", self.msg.get("request_id"))
            # 定义原视频、AI视频保存名称
            randomStr = str(uuid.uuid1().hex)
            aiFilePath = os.path.join(self.content["video"]["file_path"],
                                      randomStr + "_ai_" + self.msg.get("request_id") + ".mp4")
            logger.info("开始初始化cv2")
            if self.content["aliyun"]["vod"]["host_address"] in self.msg.get('original_url'):
                original_url = self.msg.get('original_url')
            else:
                original_url = "{}{}".format(self.content["aliyun"]["vod"]["host_address"],
                                             self.msg.get('original_url'))
            cv2tool = Cv2Util(original_url, aiFilePath=aiFilePath)
            cv2tool.build_cv2(False)
            cv2tool.build_write(False)
            logger.info("初始化cv2完成, requestId:{}", self.msg.get("request_id"))
            # cv2重试初始化次数
            cv2_init_num = 1
            # 当时视频帧数
            current_frame = 1
            high_score_image = {}
            step = int(self.content["service"]["frame_step"])
            while True:
                end_time = time.time()
                create_task_time = int(end_time - start_time)
                if create_task_time > int(self.content["service"]["timeout"]):
                    logger.info("分析超时，分析超时时间: {}s, requestId: {}", create_task_time, self.msg.get("request_id"))
                    raise ServiceException(ExceptionType.ANALYSE_TIMEOUT_EXCEPTION.value[0],
                                           ExceptionType.ANALYSE_TIMEOUT_EXCEPTION.value[1])
                if not commonProcess.is_alive():
                    logger.info("图片上传、心跳、问题反馈进程异常停止, requestId: {}", self.msg.get("request_id"))
                    raise Exception("图片上传、心跳、问题反馈进程异常停止")
                eBody = self.getEvent()
                if eBody is not None and len(eBody) > 0:
                    logger.info("eBody:{} 任务停止分析, requestId: {}", eBody, self.msg.get("request_id"))
                    cmdStr = eBody.get("command")
                    logger.info("cmdStr:{} 任务停止分析, requestId: {}", cmdStr, self.msg.get("request_id"))
                    # 接收到停止指令
                    if 'stop' == cmdStr:
                        logger.info("{} 任务停止分析", self.msg.get("request_id"))
                        self.stop_task(cv2tool, aiFilePath, commonProcess, AnalysisStatus.SUCCESS.value)
                        break
                # 检测是否断流
                if cv2tool.cap is None or not cv2tool.cap.isOpened():
                    logger.info("cv2初始化重试：{}次, requestId: {}", cv2_init_num, self.msg.get("request_id"))
                    if cv2_init_num >= 3:
                        raise ServiceException(ExceptionType.OR_VIDEO_ADDRESS_EXCEPTION.value[0],
                                               ExceptionType.OR_VIDEO_ADDRESS_EXCEPTION.value[1])
                    cv2_init_num += 1
                    time.sleep(2)
                    cv2tool.build_cv2(False)
                    continue
                is_opened, frame = cv2tool.getCap().read()
                if is_opened is None or not is_opened:
                    logger.info("总帧数: {}, 当前帧数: {}, requestId: {}", float(cv2tool.cap.get(7)), cv2tool.cap.get(1),
                                self.msg.get("request_id"))
                    if float(cv2tool.cap.get(1)) <= (float(cv2tool.cap.get(7)) - 5):
                        logger.info("离线异常结束：requestId: {}", self.msg.get("request_id"))
                        self.stop_task(cv2tool, aiFilePath, commonProcess, AnalysisStatus.TIMEOUT.value)
                        break
                    logger.info("{} 任务停止分析", self.msg.get("request_id"))
                    self.stop_task(cv2tool, aiFilePath, commonProcess, AnalysisStatus.SUCCESS.value)
                    break
                # time00 = time.time()
                # 调用AI模型
                p_result, timeOut = mod.process(copy.deepcopy(frame))
                # time11 = time.time()
                # logger.info("算法模型调度时间：{}s", int(time11-time00))
                # 原视频保存本地、AI视频保存本地
                try:
                    frame_merge = cv2tool.video_merge(copy.deepcopy(frame), copy.deepcopy(p_result[1]))
                    cv2tool.getAiVideoFile().write(frame_merge)
                except Exception as e:
                    logger.exception(e)
                # # 问题图片加入队列, 暂时写死，后期修改为真实问题
                progress = str(format(float(cv2tool.cap.get(1)) / float(cv2tool.cap.get(7)), '.4f'))
                if create_task_time % 60 == 0:
                    hbQueue.put(progress)
                if p_result[2] is not None and len(p_result[2]) > 0:
                    ai_analyse_results = p_result[2]
                    for ai_analyse_result in ai_analyse_results:
                        order = str(int(ai_analyse_result[0]))
                        high_result = high_score_image.get(order)
                        conf_c = ai_analyse_result[5]
                        if high_result is None:
                            high_score_image[order] = {
                                "or_frame": frame,
                                "ai_frame": p_result[1],
                                "current_frame": current_frame,
                                "last_frame": current_frame + step,
                                "progress": "",
                                "mode_service": "offline",
                                "model_type_code": model_type_code,
                                "model_detection_code": order,
                                "socre": conf_c
                            }
                        else:
                            if conf_c > high_result.get("socre"):
                                high_score_image[order] = {
                                    "or_frame": frame,
                                    "ai_frame": p_result[1],
                                    "current_frame": current_frame,
                                    "last_frame": current_frame + step,
                                    "progress": "",
                                    "mode_service": "offline",
                                    "model_type_code": model_type_code,
                                    "model_detection_code": order,
                                    "socre": conf_c
                                }
                if current_frame % int(self.content["service"]["frame_step"]) == 0 and len(high_score_image) > 0:
                    for value in high_score_image.values():
                        self.imageQueue.put(value)
                    high_score_image.clear()
                current_frame += 1
            logger.info("实时进程任务完成，requestId:{}", self.msg.get("request_id"))
        except ServiceException as s:
            commonProcess.sendEvent({"command": "stop_heartbeat_imageFileUpdate"})
            logger.error("服务异常，异常编号:{}, 异常描述：{}, requestId:{}", s.code, s.msg, self.msg.get("request_id"))
            self.sendResult(message_feedback(self.msg.get("request_id"), AnalysisStatus.FAILED.value,
                                             AnalysisType.OFFLINE.value,
                                             s.code,
                                             s.msg,
                                             analyse_time=TimeUtils.now_date_to_str()))
        except Exception as e:
            logger.error("requestId:{}, 服务异常：", self.msg.get("request_id"))
            logger.exception(e)
            self.sendResult(message_feedback(self.msg.get("request_id"), AnalysisStatus.FAILED.value,
                                             AnalysisType.OFFLINE.value,
                                             ExceptionType.SERVICE_INNER_EXCEPTION.value[0],
                                             ExceptionType.SERVICE_INNER_EXCEPTION.value[1],
                                             analyse_time=TimeUtils.now_date_to_str()))
        finally:
            if cv2tool is not None:
                cv2tool.stop_cv2()
            time.sleep(3)
            commonProcess.sendEvent({"command": "stop"})
            commonProcess.join()
            # 删除本地视频文件
            if aiFilePath is not None and os.path.exists(aiFilePath):
                logger.info("开始删除AI视频, aiFilePath: {}, requestId: {}", aiFilePath, self.msg.get("request_id"))
                os.remove(aiFilePath)
                logger.info("删除AI视频成功, aiFilePath: {}, requestId: {}", aiFilePath, self.msg.get("request_id"))


class PhotosIntelligentRecognitionProcess(IntelligentRecognitionProcess):
    pass
