# -*- coding: utf-8 -*-
import cv2
import subprocess as sp
import numpy as np
from loguru import logger
from exception.CustomerException import ServiceException
from enums.ExceptionEnum import ExceptionType


class Cv2Util():

    def __init__(self, pullUrl, pushUrl=None, orFilePath=None, aiFilePath=None):
        self.pullUrl = pullUrl
        self.pushUrl = pushUrl
        self.orFilePath = orFilePath
        self.aiFilePath = aiFilePath
        self.cap = None
        self.p = None
        self.or_video_file = None
        self.ai_video_file = None
        self.fps = None
        self.width = None
        self.height = None

    def stop_cv2(self):
        if self.p:
            logger.info("关闭管道")
            self.p.stdin.close()
            self.p.terminate()
            self.p.wait()
            logger.info("关闭管道完成")
        if self.or_video_file:
            logger.info("关闭原视频写入流")
            self.or_video_file.release()
            logger.info("关闭原视频写入流完成")
        if self.ai_video_file:
            logger.info("关闭AI视频写入流")
            self.ai_video_file.release()
            logger.info("关闭AI视频写入流完成")
        if self.cap:
            logger.info("关闭cv2 cap")
            self.cap.release()
            logger.info("关闭cv2 cap完成")

    # 构建 cv2
    def build_cv2(self, is_online=True):
        try:
            if self.cap is not None:
                logger.info("关闭cap")
                self.cap.release()
            if self.p is not None:
                logger.info("关闭管道")
                self.p.stdin.close()
                self.p.terminate()
                self.p.wait()
            if self.pullUrl is None:
                logger.error("拉流地址不能为空！")
                raise ServiceException(ExceptionType.PULL_STREAM_URL_EXCEPTION.value[0],
                                       ExceptionType.PULL_STREAM_URL_EXCEPTION.value[1])
            self.cap = cv2.VideoCapture(self.pullUrl)
            if self.fps is None or self.fps == 0:
                self.fps = int(self.cap.get(cv2.CAP_PROP_FPS))
            if self.width is None or self.width == 0:
                self.width = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
            if self.height is None or self.height == 0:
                self.height = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
            if is_online:
                if self.pushUrl is None:
                    logger.error("推流地址不能为空！")
                    raise ServiceException(ExceptionType.PUSH_STREAM_URL_EXCEPTION.value[0],
                                           ExceptionType.PUSH_STREAM_URL_EXCEPTION.value[1])
                command = ['/usr/bin/ffmpeg',
                           '-y', # 不经过确认，输出时直接覆盖同名文件。
                           '-f', 'rawvideo',
                           '-vcodec','rawvideo',
                           '-pix_fmt', 'bgr24',
                           '-s', "{}x{}".format(self.width, self.height),
                           '-r', str(self.fps),
                           '-i', '-', # 指定输入文件
                           '-c:v', 'libx264', # 指定视频编码器
                           '-pix_fmt', 'yuv420p',
                           '-preset', 'ultrafast', # 指定输出的视频质量，会影响文件的生成速度，有以下几个可用的值 ultrafast,
                           # superfast, veryfast, faster, fast, medium, slow, slower, veryslow。
                           '-f', 'flv',
                           self.pushUrl]
                # 管道配置
                logger.info("fps:{}|height:{}|width:{}", self.fps, self.height, self.width)
                self.p = sp.Popen(command, stdin=sp.PIPE)
        except ServiceException as s:
            raise s
        except Exception as e:
            logger.error("初始化cv2异常：")
            logger.exception(e)


    def build_write(self, is_online=True):
        try:
            if self.fps is None or self.width is None or self.height is None or self.aiFilePath is None:
                logger.error("aiFilePath配置不能为空，请先执行build_cv2的方法，再执行次方法！")
                raise ServiceException(ExceptionType.AI_VIDEO_ADDRESS_EXCEPTION.value[0],
                                       ExceptionType.AI_VIDEO_ADDRESS_EXCEPTION.value[1])
            if is_online:
                if self.orFilePath is None:
                    logger.error("pushUrl推流地址和orFilePath原视频地址都不能为空！")
                    raise ServiceException(ExceptionType.OR_VIDEO_ADDRESS_EXCEPTION.value[0],
                                           ExceptionType.OR_VIDEO_ADDRESS_EXCEPTION.value[1])

                self.or_video_file = cv2.VideoWriter(self.orFilePath, cv2.VideoWriter_fourcc(*'mp4v'), self.fps, (self.width, self.height))
            self.ai_video_file = cv2.VideoWriter(self.aiFilePath, cv2.VideoWriter_fourcc(*'mp4v'), self.fps, (self.width * 2, self.height))
        except ServiceException as s:
            raise s
        except Exception as e:
            logger.error("初始化管道失败：")
            logger.exception(e)

    def video_merge(self, frame1, frame2):
        frameLeft = cv2.resize(frame1, (int(self.width), int(self.height)), interpolation=cv2.INTER_LINEAR)
        frameRight = cv2.resize(frame2, (int(self.width), int(self.height)), interpolation=cv2.INTER_LINEAR)
        frame_merge = np.hstack((frameLeft, frameRight))
        return frame_merge

    def getP(self):
        if self.p is None:
            logger.error("获取管道为空！")
            raise ServiceException(ExceptionType.PULL_PIPELINE_INIT_EXCEPTION.value[0],
                                   ExceptionType.PULL_PIPELINE_INIT_EXCEPTION.value[1])
        return self.p

    def getCap(self):
        if self.cap is None:
            logger.error("获取cv2为空！")
            raise ServiceException(ExceptionType.CV2_IS_NULL_EXCEPTION.value[0],
                                   ExceptionType.CV2_IS_NULL_EXCEPTION.value[1])
        return self.cap

    def getOrVideoFile(self):
        if self.or_video_file is None:
            logger.error("获取原视频写入对象为空！")
            raise ServiceException(ExceptionType.OR_WRITE_OBJECT_EXCEPTION.value[0],
                                   ExceptionType.OR_WRITE_OBJECT_EXCEPTION.value[1])
        return self.or_video_file

    def getAiVideoFile(self):
        if self.ai_video_file is None:
            logger.error("获取AI视频写入对象为空！")
            raise ServiceException(ExceptionType.AI_WRITE_OBJECT_EXCEPTION.value[0],
                                   ExceptionType.AI_WRITE_OBJECT_EXCEPTION.value[1])
        return self.ai_video_file