from kafka import KafkaProducer, KafkaConsumer, TopicPartition, OffsetAndMetadata
from kafka.errors import kafka_errors
import json
from loguru import logger


# 生产者
class CustomerKafkaProducer():

    def __init__(self, content):
        self.content = content
        configs = self.content["kafka"][self.content["dsp"]["active"]]["producer"]
        self.customerProducer = KafkaProducer(bootstrap_servers=self.content["kafka"][self.content["dsp"]["active"]]["bootstrap_servers"],
                                              acks=configs["acks"],
                                              retries=configs["retries"],
                                              linger_ms=configs["linger_ms"],
                                              retry_backoff_ms=configs["retry_backoff_ms"],
                                              max_in_flight_requests_per_connection=configs[
                                                  "max_in_flight_requests_per_connection"],
                                              key_serializer=lambda m: json.dumps(m).encode('utf-8'),
                                              value_serializer=lambda m: json.dumps(m).encode('utf-8'))

    # 获取kafka生产者
    def get_producer(self):
        if self.customerProducer:
            return self.customerProducer
        logger.info("配置kafka生产者")
        configs = self.content["kafka"][self.content["dsp"]["active"]]["producer"]
        self.customerProducer = KafkaProducer(bootstrap_servers=self.content["kafka"][self.content["dsp"]["active"]]["bootstrap_servers"],
                                              acks=configs["acks"],
                                              retries=configs["retries"],
                                              linger_ms=configs["linger_ms"],
                                              retry_backoff_ms=configs["retry_backoff_ms"],
                                              max_in_flight_requests_per_connection=configs[
                                                  "max_in_flight_requests_per_connection"],
                                              key_serializer=lambda m: json.dumps(m).encode('utf-8'),
                                              value_serializer=lambda m: json.dumps(m).encode('utf-8'))
        logger.info("配置生产者完成")
        return self.customerProducer

    # mode 模式1：异步发送 2：同步发送
    # def on_send_success(record_metadata): 成功回调
    # def on_send_error(excp): 失败回调
    def sender(self, topic, key, message, mode, customer_send_success=None, customer_send_error=None):
        logger.info("kafka发送信息，topic:{}|key:{}|message:{}|mode:{}", topic, key, message, mode)
        if mode == 1:
            if not customer_send_success:
                customer_send_success = CustomerKafkaProducer.on_send_success
            if not customer_send_error:
                customer_send_error = CustomerKafkaProducer.on_send_error
            self.customerProducer.send(topic=topic, key=key, value=message).add_callback(
                customer_send_success).add_errback(customer_send_error)
        if mode == 2:
            try:
                self.customerProducer.send(topic=topic, key=key, value=message).get(timeout=30)
                logger.info("kafka同步发送信息成功")
            except kafka_errors as e:
                logger.error("kafka同步发送消息异常：")
                logger.exception(e)
                raise e

    def close_producer(self):
        self.customerProducer.flush()
        self.customerProducer.close()
        logger.info("kafka生产者关闭完成")

    def on_send_success(record_metadata):
        logger.info("kafka异步发送信息成功，topic:{}|partition:{}|offset:{}", record_metadata.topic, record_metadata.partition,
                    record_metadata.offset)

    def on_send_error(excp):
        logger.error("kafka异步发送消息异常：")
        logger.error(excp)


# 生产者
class CustomerKafkaConsumer():

    def __init__(self, content):
        logger.info("初始化消费者")
        self.content = content
        configs = self.content["kafka"][self.content["dsp"]["active"]]["consumer"]
        self.customerConsumer = KafkaConsumer(bootstrap_servers=self.content["kafka"][self.content["dsp"]["active"]]["bootstrap_servers"],
                                              client_id=configs["client_id"],
                                              group_id=configs["group_id"],
                                              auto_offset_reset=configs["auto_offset_reset"],
                                              enable_auto_commit=configs["enable_auto_commit"],
                                              max_poll_records=configs["max_poll_records"],
                                              value_deserializer=lambda m: json.loads(m.decode('utf-8')))
        logger.info("初始化消费者完成")

    def get_consumer(self):
        logger.info("获取消费者！")
        if self.customerConsumer:
            logger.info("获取消费者成功！")
            return self.customerConsumer
        configs = self.content["kafka"][self.content["dsp"]["active"]]["consumer"]
        self.customerConsumer = KafkaConsumer(bootstrap_servers=self.content["kafka"][self.content["dsp"]["active"]]["bootstrap_servers"],
                                              client_id=configs["client_id"],
                                              group_id=configs["group_id"],
                                              auto_offset_reset=configs["auto_offset_reset"],
                                              enable_auto_commit=configs["enable_auto_commit"],
                                              max_poll_records=configs["max_poll_records"],
                                              value_deserializer=lambda m: json.loads(m.decode('utf-8')))
        logger.info("消费者获取成功")
        return self.customerConsumer

    def subscribe(self, topics=()):
        logger.info("kafka生产者订阅topic:{}", topics)
        if topics is None or len(topics) == 0:
            logger.error("消费者订阅topic不能为空！")
            raise Exception("消费者订阅topic不能为空！")
        # 手动配置分区
        # customer_partition = []
        # for topic in topics:
        #     for p in self.content["kafka"]["topic"][topic]["partition"]:
        #         customer_partition.append(TopicPartition(topic, p))
        # self.customerConsumer.assign(customer_partition)
        # 自动配置
        self.customerConsumer.subscribe(topics=topics)
        logger.info("kafka生产者订阅topic完成")

    def commit_offset(self, message):
        logger.info("消费者开始提交offset,topic:{}|offset:{}|partition:{}", message.topic, message.offset + 1,
                    message.partition)
        tp = TopicPartition(topic=message.topic, partition=message.partition)
        self.customerConsumer.commit(offsets={tp: (OffsetAndMetadata(message.offset + 1, None))})
        logger.info("消费者提交offset完成,topic:{}|offset:{}|partition:{}", message.topic, message.offset + 1,
                    message.partition)
